use std::time::Instant;

#[derive(Copy, Clone, Debug)]
struct Bitset {
    set: u16,
}

impl Bitset {
    pub fn new() -> Self {
        Self { set: 0 }
    }

    #[inline(always)]
    fn get(&self, nr: usize) -> bool {
        self.set & (1 << nr) != 0
    }

    #[inline(always)]
    fn set(&mut self, nr: usize) {
        self.set |= 1 << nr;
    }

    #[inline(always)]
    fn unset(&mut self, nr: usize) {
        self.set &= !(1 << nr);
    }
}

const BLOCK_SIZE: usize = 4;
//const MAX_BLOCK_SIZE: usize = 10;

struct SField {
    field: Vec<usize>,
    fixed: Vec<usize>,
    size: usize,
    pos: usize,
    pos_last_placed: usize,
    num_fields: usize,
    possible_values: Vec<Vec<usize>>,

    // For each row/col/block we create a bitset that represents the numbers
    // that are already in that row, col or block.
    // This way we can check very fast if a nr fits into a specific field or not.
    rows: [Bitset; BLOCK_SIZE * BLOCK_SIZE],
    cols: [Bitset; BLOCK_SIZE * BLOCK_SIZE],
    blocks: [Bitset; BLOCK_SIZE * BLOCK_SIZE],
}

#[inline(always)]
fn xy_to_pos(x: usize, y: usize, size: usize) -> usize {
    y * size + x
}

#[inline(always)]
fn pos_to_xy_blocknr(pos: usize, size: usize) -> (usize, usize, usize) {
    let x = pos % size;
    let y = pos / size;

    let block_x = x / BLOCK_SIZE;
    let block_y = y / BLOCK_SIZE;

    let block_nr = block_y * BLOCK_SIZE + block_x;

    (x, y, block_nr)
}

impl SField {
    pub fn new() -> SField {
        const SIZE: usize = BLOCK_SIZE * BLOCK_SIZE;

        // EMPTY FIELD
        // let field = vec![0; size * size];

        #[rustfmt::skip]
        let field = match BLOCK_SIZE {
            4 => vec![
                0usize,15,0,0,0,0,11,4,0,5,0,0,0,0,12,8,
                12,0,0,9,0,1,0,5,0,0,8,15,0,0,0,13,
                0,0,3,0,0,10,13,0,0,11,4,0,15,0,6,0,
                13,10,0,11,0,0,0,14,0,3,2,0,0,9,0,0,
                0,0,15,10,8,0,0,0,0,0,14,0,0,6,2,0,
                0,0,14,0,0,0,6,0,0,0,0,0,1,0,0,0,
                0,11,0,8,3,0,15,1,6,0,0,0,0,0,0,7,
                0,6,0,7,0,0,0,0,8,13,0,0,10,0,0,0,
                0,14,0,2,0,0,0,0,3,0,5,0,11,15,9,0,
                0,0,0,12,11,0,2,0,0,0,0,0,0,0,0,0,
                0,8,0,0,6,14,1,0,13,15,0,0,0,0,0,10,
                10,0,0,3,9,0,7,0,0,1,0,0,13,12,8,0,
                7,0,0,0,0,0,14,0,0,0,0,0,0,0,0,9,
                0,3,0,14,0,0,9,6,0,0,0,0,0,4,0,0,
                0,4,6,0,0,7,0,0,0,0,0,0,0,0,0,0,
                5,2,0,0,0,4,10,15,1,0,3,0,0,0,7,0,
            ],
            3 => vec![
                0,0,0,0,0,0,0,0,9,
                0,0,0,0,8,9,0,2,0,
                0,0,0,0,2,0,4,0,0,
                0,0,4,0,6,0,0,0,8,
                0,0,0,5,0,0,0,0,0,
                0,6,5,0,0,2,0,7,4,
                0,3,0,0,0,5,0,4,0,
                0,0,1,8,0,0,0,0,0,
                0,0,8,2,0,0,0,6,0,
            ],
            _ => panic!("Unsupported block size. Only 3 and 4 is ok for now.")
        };

        // MANY SOLUTIONS
        // #[rustfmt::skip]
        // let field = vec![
        //     0,0,0,0,0,0,0,0,9,
        //     0,0,0,0,8,9,0,2,0,
        //     0,0,0,0,2,0,4,0,0,
        //     0,0,4,0,6,0,0,0,8,
        //     0,0,0,5,0,0,0,0,0,
        //     0,6,5,0,0,2,0,7,4,
        //     0,3,0,0,0,5,0,4,0,
        //     0,0,1,8,0,0,0,0,0,
        //     0,0,8,2,0,0,0,6,0,
        // ];

        // HARD
        // #[rustfmt::skip]
        // let field = vec![
        //     8,4,0,0,6,0,5,0,1,
        //     0,0,0,0,0,3,0,4,0,
        //     0,0,6,9,0,0,0,0,7,
        //     0,2,0,7,1,0,0,0,6,
        //     0,0,0,6,3,0,0,0,0,
        //     9,0,0,0,0,0,0,5,0,
        //     0,0,0,0,4,0,0,6,0,
        //     2,0,0,0,0,0,1,8,0,
        //     0,0,5,0,0,0,3,0,0,
        // ];

        // EASY
        // #[rustfmt::skip]
        // let field = vec![
        //     5,9,0,6,1,3,0,0,0,
        //     0,0,0,9,0,0,5,0,0,
        //     8,0,3,0,5,7,6,4,0,
        //     0,7,5,0,0,0,4,0,6,
        //     0,6,0,7,4,0,2,0,8,
        //     2,0,8,0,0,0,7,5,3,
        //     0,0,0,5,6,1,0,7,0,
        //     0,0,1,0,7,0,9,0,0,
        //     7,3,6,0,2,0,0,0,0,
        // ];

        let mut fixed = Vec::with_capacity(SIZE * SIZE);
        field.iter().for_each(|e| {
            if *e == 0 {
                fixed.push(0);
            } else {
                fixed.push(1);
            }
        });

        // create bitsets
        // rows
        let mut bit_rows = [Bitset::new(); SIZE];
        for row_nr in 0..SIZE {
            let mut row = [0; SIZE];
            for count in 0..SIZE {
                row[count] = field[xy_to_pos(count, row_nr, SIZE)];
            }

            row.into_iter()
                .filter(|val| *val != 0)
                .for_each(|val| bit_rows[row_nr].set(val - 1));
        }

        // cols
        let mut bit_cols = [Bitset::new(); SIZE];
        for col_nr in 0..SIZE {
            let mut col = [0; SIZE];
            for count in 0..SIZE {
                col[count] = field[xy_to_pos(col_nr, count, SIZE)];
            }

            col.into_iter()
                .filter(|val| *val != 0)
                .for_each(|val| bit_cols[col_nr].set(val - 1));
        }

        // blocks
        let mut bit_blocks = [Bitset::new(); SIZE];

        for block_nr in 0..SIZE {
            let block_start_x = block_nr * BLOCK_SIZE % SIZE;
            let block_start_y = block_nr * BLOCK_SIZE / SIZE * BLOCK_SIZE;

            let mut block = [0; SIZE];
            for count in 0..SIZE {
                let block_offset_x = count % BLOCK_SIZE;
                let block_offset_y = count / BLOCK_SIZE;

                let field_x = block_start_x + block_offset_x;
                let field_y = block_start_y + block_offset_y;

                //dbg!(
                //block_nr,
                //block_start_x,
                //block_start_y,
                //block_offset_x,
                //block_offset_y,
                //field_x,
                //field_y,
                //);

                block[count] = field[xy_to_pos(field_x, field_y, SIZE)];
            }

            block
                .into_iter()
                .filter(|x| *x != 0)
                .for_each(|x| bit_blocks[block_nr].set(x - 1));
        }

        //dbg!(bit_rows);
        //dbg!(bit_cols);
        //dbg!(bit_blocks);

        SField {
            field,
            fixed,
            size: SIZE,
            pos: 0,
            pos_last_placed: 0,
            num_fields: SIZE * SIZE,
            possible_values: vec![vec![]; SIZE * SIZE],
            rows: bit_rows,
            cols: bit_cols,
            blocks: bit_blocks,
        }
    }

    fn build_possible_values_db(&mut self) {
        for idx in 0..self.num_fields {
            if self.fixed[idx] == 1 {
                continue;
            }
            self.pos = idx;
            // try all values between 1 and =self.size and remember the good ones
            let mut good_ones = Vec::with_capacity(self.size);
            for nr in 1..=self.size {
                if self.ok(nr) {
                    good_ones.push(nr);
                }
            }
            self.possible_values[idx] = good_ones;
        }
        self.pos = 0;
    }

    pub fn solve_backtracking(&mut self) -> bool {
        let mut found_solution = false;
        let mut num_solutions = 0;

        let mut loop_count = 0;
        let mut last_loop_count = 0;
        let mut now = Instant::now();
        loop {
            //std::thread::sleep_ms(30);
            //self.print_clear();
            //self.print();
            loop_count += 1;
            if now.elapsed().as_millis() >= 1000 {
                now = Instant::now();
                self.print_clear();
                self.print();
                println!("{} loops/s", loop_count - last_loop_count);
                last_loop_count = loop_count;
            }
            if self.is_end() {
                found_solution = true;
                num_solutions += 1;
                //if num_solutions % 100 == 0 {
                //self.print_clear();
                //println!("Solutions: {}", num_solutions);
                //self.print();
                //}

                if self.is_fixed() {
                    continue;
                }
                self.clear_current_field();
                self.prev();
                self.goto_prev_free_field();
            }

            if self.is_fixed() {
                self.next();
                continue;
            }

            if self.goto_next_free_field() {
                if self.put_valid_nr() {
                    self.next();
                    continue;
                } else {
                    //println!("put_valid_nr failed for pos {}", self.pos);
                    //std::thread::sleep_ms(300);
                    self.clear_current_field();
                    if !self.prev() {
                        println!("Number of solutions: {}", num_solutions);
                        break;
                    }
                    if !self.goto_prev_free_field() {
                        println!("Number of solutions: {}", num_solutions);
                        break;
                    }
                }
            }
        }

        found_solution
    }

    #[inline(always)]
    fn print_gray(&self) {
        print!("\x1b\x5b\x31\x3b\x33\x30\x6d");
    }

    #[inline(always)]
    fn print_green(&self) {
        print!("\x1b\x5b\x31\x3b\x33\x32\x6d");
    }

    #[inline(always)]
    fn print_red(&self) {
        print!("\x1b\x5b\x31\x3b\x33\x31\x6d");
    }

    #[inline(always)]
    fn print_neutral(&self) {
        print!("\x1b\x5b\x31\x3b\x30\x6d");
    }

    #[inline(always)]
    fn print_clear(&self) {
        print!("\x1b\x5b\x48\x1b\x5b\x32\x4a");
    }

    #[inline(always)]
    fn print(&self) {
        for i in 0..self.num_fields {
            if i != 0 && i % self.size == 0 {
                println!();
            }

            if i == self.pos_last_placed {
                self.print_red();
            } else if i == self.pos {
                self.print_green();
            } else if self.get_field_at_pos(i) == 0 {
                self.print_gray();
            }

            print!("{:2} ", self.get_field_at_pos(i));

            if i == self.pos || i == self.pos_last_placed || self.get_field_at_pos(i) == 0 {
                self.print_neutral();
            }
        }
        println!();
    }

    #[inline(always)]
    fn clear_current_field(&mut self) {
        self.set(0);
    }

    #[inline(always)]
    fn goto_prev_free_field(&mut self) -> bool {
        while {
            if !self.is_fixed() {
                return true;
            }
            self.prev()
        } {}

        false
    }

    #[inline(always)]
    fn goto_next_free_field(&mut self) -> bool {
        while {
            if !self.is_fixed() {
                return true;
            }
            self.next()
        } {}

        false
    }

    #[inline(always)]
    fn _put_valid_nr(&mut self) -> bool {
        let current_nr = self.get_field_at_pos(self.pos);

        for nr in current_nr..self.size + 1 {
            if self.ok(nr) {
                self.set(nr);
                return true;
            }
        }

        false
    }

    #[inline(always)]
    fn put_valid_nr(&mut self) -> bool {
        let current_nr = self.get_field_at_pos(self.pos);

        // safety:
        //     self.possible_vals is initialized with self.size * self.size
        //     so self.pos can safely be used to index here
        let possible_vals = unsafe { self.possible_values.get_unchecked(self.pos) };

        for nr in possible_vals.iter() {
            if *nr <= current_nr {
                continue;
            }

            let nr = *nr;
            if self.ok(nr) {
                self.set(nr);
                return true;
            }
        }
        false
    }

    #[inline(always)]
    fn is_end(&self) -> bool {
        !self.field.contains(&0)
    }

    #[inline(always)]
    fn ok(&self, nr: usize) -> bool {
        //self.block_ok(nr) && self.row_ok(nr) && self.col_ok(nr)

        let (x, y, block_nr) = pos_to_xy_blocknr(self.pos, self.size);

        !self.rows[y].get(nr - 1) && !self.cols[x].get(nr - 1) && !self.blocks[block_nr].get(nr - 1)
    }

    #[inline(always)]
    fn is_fixed(&self) -> bool {
        // safety: self.pos can be used to index the field unchecked
        //         since the only methods modifying self.pos are
        //         `next()` and `prev()` and they do bounds checking
        unsafe { *self.fixed.get_unchecked(self.pos) == 1 }
    }

    #[inline(always)]
    fn get_field_at_pos(&self, pos: usize) -> usize {
        // safety:
        //     TODO
        //     would need to mathematically explain that the calculations in
        //     get_row(), get_col() and get_block() can never exceed 0..self.num_fields
        unsafe { *self.field.get_unchecked(pos) }
    }

    #[inline(always)]
    fn set(&mut self, nr: usize) {
        let old = unsafe { self.field.get_unchecked_mut(self.pos) };

        // if another number > 0 was in that cell, we need to remove that number from our bitsets
        if *old > 0 {
            let (x, y, block_nr) = pos_to_xy_blocknr(self.pos, self.size);
            self.rows[y].unset(*old - 1);
            self.cols[x].unset(*old - 1);
            self.blocks[block_nr].unset(*old - 1);
        }

        // add new nr to our bitsets if it is a meaningful (> 0) number
        if nr > 0 {
            let (x, y, block_nr) = pos_to_xy_blocknr(self.pos, self.size);
            self.rows[y].set(nr - 1);
            self.cols[x].set(nr - 1);
            self.blocks[block_nr].set(nr - 1);
        }

        // write the new number into the field
        *old = nr;

        // remember the last placed position for debugging purposes
        self.pos_last_placed = self.pos;
    }

    /*
    #[inline(always)]
    fn get_row(&self, row: &mut [usize]) {
        for (idx, row_elem) in row.iter_mut().enumerate() {
            *row_elem = self.get_field_at_pos((self.pos / self.size) * self.size + idx);
        }
    }

    #[inline(always)]
    fn get_col(&self, col: &mut [usize]) {
        for (idx, col_elem) in col.iter_mut().enumerate() {
            *col_elem = self.get_field_at_pos(idx * self.size + self.pos % self.size);
        }
    }

    #[inline(always)]
    fn get_block(&self, block: &mut [usize]) {
        let block_start_row = self.pos / self.size / self.block_size * self.block_size;
        let block_start_col = self.pos % self.size / self.block_size * self.block_size;

        for r in 0..self.block_size {
            for c in 0..self.block_size {
                block[r * self.block_size + c] =
                    self.get_field_at_pos(self.size * (block_start_row + r) + block_start_col + c);
            }
        }
    }

    #[inline(always)]
    fn block_ok(&self, nr: usize) -> bool {
        let mut block = [0; MAX_BLOCK_SIZE * MAX_BLOCK_SIZE];
        self.get_block(&mut block[0..self.size]);
        !block.contains(&nr)
    }

    #[inline(always)]
    fn row_ok(&self, nr: usize) -> bool {
        let mut row = [0; MAX_BLOCK_SIZE * MAX_BLOCK_SIZE];
        self.get_row(&mut row[0..self.size]);
        !row.contains(&nr)
    }

    #[inline(always)]
    fn col_ok(&self, nr: usize) -> bool {
        let mut col = [0; MAX_BLOCK_SIZE * MAX_BLOCK_SIZE];
        self.get_col(&mut col[0..self.size]);
        !col.contains(&nr)
    }
    */

    #[inline(always)]
    fn next(&mut self) -> bool {
        if self.pos == self.num_fields - 1 {
            return false;
        }

        //println!("next {} -> {}", self.pos, self.pos + 1);
        self.pos += 1;
        true
    }

    #[inline(always)]
    fn prev(&mut self) -> bool {
        if self.pos == 0 {
            return false;
        }

        self.pos -= 1;
        true
    }

    #[inline(always)]
    fn solve(&mut self) {
        let now = Instant::now();
        if !self.solve_backtracking() {
            println!("there is no solution.");
        }
        println!("took {:.3}s", now.elapsed().as_secs_f64());
    }
}

fn run() -> Result<(), String> {
    let mut field = SField::new();

    field.build_possible_values_db();

    field.solve();
    Ok(())
}

fn main() {
    if let Err(e) = run() {
        println!("{}", e)
    }
}
